#!/bin/env python3
"""Stupid Inefficient Exhaustive Shaping Test Architecture."""
import argparse
import json
import os
import random
import sys
import itertools
import re

import fontFeatures
import tqdm
from babelfont import load
from fontFeatures.shaperLib.Buffer import Buffer
from fontFeatures.shaperLib.Shaper import Shaper
from fontFeatures.ttLib import unparse
from fontTools.ttLib import TTFont
from vharfbuzz import Vharfbuzz

def main(args=None):
    parser = argparse.ArgumentParser(description="Test coverage of OT features.")
    subparsers = parser.add_subparsers(required=True, dest="cmd")
    report = subparsers.add_parser("report")
    report.add_argument("--suite", help="Test suite to read", metavar="JSON", required=True)
    report.add_argument("font", metavar="TTF", help="font file")

    generate = subparsers.add_parser("generate")
    generate.add_argument(
        "--suite",
        help="Test suite to update/create (defaults to test.json)",
        metavar="JSON",
        default="test.json",
    )
    generate.add_argument(
        "--language",
        help="Language to pass to shaper",
    )
    generate.add_argument(
        "--only", action="store_true", help="Add \"'only': FONT\" to JSON test cases"
    )
    generate.add_argument(
        "--restrict-unicodes",
        help="Restrict unicodes to the given range(s, comma separated)",
        metavar="RANGE",
    )
    generate.add_argument(
        "--just-blobs",
        action="store_true",
        help="Only use glyph sequences known to trigger rules",
    )
    generate.add_argument(
        "--no-dotted-circle",
        action="store_true",
        help="Don't add tests with dotted circles",
    )
    generate.add_argument(
        "--string-length",
        type=int,
        default=5,
        help="Maximum string length",
    )
    generate.add_argument("font", metavar="TTF", help="font file")
    args = parser.parse_args(args)


    def restrict(unicodes, range_arg):
        allowable = set()
        ranges = range_arg.split(",")
        for r in ranges:
            m = re.match(r"\s*([0-9A-Fa-f]+)-([0-9A-Fa-f]+)", r)
            if not m:
                raise ValueError(f"Could not understand range {r}")
            for cp in range(int(m[1], 16), int(m[2], 16) + 1):
                allowable.add(chr(cp))
        return set(allowable) & set(unicodes)


    font = TTFont(args.font)
    ff = unparse(font)
    bbfont = load(args.font)
    num_routines = 0
    num_rules = 0


    # Monkeypatch shaperLib functions
    def wrap_do_apply(fn):
        def result(*args, **kwargs):
            fn.__self__.applied = True
            fn.__self__.parent.applied = True
            return fn(*args, **kwargs)

        return result


    for r in ff.routines:
        r.applied = False
        num_routines += 1
        for rule in r.rules:
            rule.applied = False
            rule.parent = r
            rule._do_apply = wrap_do_apply(rule._do_apply)
            num_rules += 1


    def count_triggered(ff):
        num_triggered_routines = 0
        num_triggered_rules = 0
        for r in ff.routines:
            if r.applied:
                num_triggered_routines += 1
            for rule in r.rules:
                if rule.applied:
                    num_triggered_rules += 1
        return num_triggered_routines, num_triggered_rules


    num_triggered_routines = 0
    num_triggered_rules = 0

    shaper = Shaper(ff, bbfont)

    unicodes = list(font.getBestCmap().keys())

    dotted_circle = font.getBestCmap().get(0x25CC)

    if args.cmd == "generate":
        # Remove those not involved in rules?
        # Complex shapers may mess this up?
        involved = set()
        reversecmap = font["cmap"].buildReversed()

        blobs = []
        for r in ff.routines:
            try:
                involved = involved | r.involved_glyphs
            except Exception:
                pass
            for rule in r.rules:
                if isinstance(rule, fontFeatures.Attachment):
                    fullinput = [list(rule.bases.keys())] + [list(rule.marks.keys())]
                else:
                    fullinput = (
                        rule.precontext
                        + getattr(rule, "input", [])
                        + getattr(rule, "glyphs", [])
                        + rule.postcontext
                    )
                allowable = []
                for slot in fullinput:
                    this_slot = []
                    for x in slot:
                        if x in reversecmap:
                            this_slot.extend([chr(c) for c in reversecmap[x]])
                    allowable.append(this_slot)
                if all(allowable):
                    blobs.extend(["".join(blob) for blob in itertools.product(*allowable)])

        unicodes = [c for c in unicodes if font.getBestCmap()[c] in involved]
        unicodes = [chr(c) for c in unicodes]
        if args.restrict_unicodes:
            unicodes = list(restrict(unicodes, args.restrict_unicodes))
        if args.just_blobs:
            unicodes = []
        unicodes.extend(blobs)


    rule = tqdm.tqdm(
        total=num_rules,
        position=1,
        colour="green",
        desc="Rules  ",
        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
    )
    routine = tqdm.tqdm(
        total=num_routines,
        position=2,
        colour="yellow",
        desc="Lookups",
        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
    )
    vh = Vharfbuzz(args.font)


    def run_a_test(string, tests=None, output=True):
        global num_triggered_routines
        global num_triggered_rules
        buf = Buffer(bbfont, unicodes=string)#,language=args.language)
        try:
            shaper.execute(buf)
        except Exception:
            return False

        if dotted_circle and args.no_dotted_circle:
            # Check HB, FF shaper doesn't always add dotted circles
            buf2 = vh.shape(string)
            if dotted_circle in vh.serialize_buf(buf2):
                return False

        new_triggered_routines, new_triggered_rules = count_triggered(ff)
        if (
            new_triggered_routines > num_triggered_routines
            or new_triggered_rules > num_triggered_rules
        ):
            if output:
                routine.write(string)
            if tests is not None:
                tests.append({"input": string})
            rule.n = 0
            rule.update(new_triggered_rules)
            routine.n = 0
            routine.update(new_triggered_routines)
        num_triggered_rules = new_triggered_rules
        num_triggered_routines = new_triggered_routines
        return new_triggered_rules == num_rules  # Done?


    tests = []
    suite = {"tests": []}
    if args.suite and os.path.exists(args.suite):
        suite = json.load(open(args.suite))
        if "tests" not in suite:
            suite["tests"] = []
        for t in suite["tests"]:
            if "only" in t and os.path.basename(args.font) != t["only"]:
                continue
            run_a_test(t["input"], tests=None, output=False)

    if args.cmd == "report":
        sys.exit(0)

    new_tests = []

    try:
        while True:
            string = "".join(
                random.choices(unicodes, k=random.randint(2, args.string_length))
            )
            if run_a_test(string, tests=new_tests, output=True):
                break
    except KeyboardInterrupt as e:
        pass

    fh = open(args.suite, "w")
    params = {}
    if args.language:
        params["language"] = args.language
    # Use HB for gold-standard output
    for test in new_tests:
        buf = vh.shape(test["input"],parameters=params)
        test["expectation"] = vh.serialize_buf(buf)
        if args.only:
            test["only"] = os.path.basename(args.font)
        test["note"] = "Added by SIESTA"
        for k,v in params.items():
            test[k] = v

    suite["tests"].extend(new_tests)
    json.dump(suite, fh, sort_keys=True, indent=4, ensure_ascii=False)

    print("The following lookups were not triggered:")
    for r in ff.routines:
        if not r.applied:
            print(r.asFea())
    sys.exit(0)

if __name__ == "__main__":
    main()


#!/usr/bin/env python
import sys
from fontTools.ttLib import TTFont
from fontFeatures.ttLib import unparse
from fontFeatures.ttLib.GDEFUnparser import GDEFUnparser
from fontFeatures.optimizer import Optimizer
from argparse import ArgumentParser

import logging
import os

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logging.basicConfig(level=LOGLEVEL)

import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return "# [warning] %s\n" % (message)


warnings.formatwarning = warning_on_one_line

def main(args=None):
    parser = ArgumentParser()
    parser.add_argument("input", help="font file to process", metavar="FILE")
    parser.add_argument(
        "--gdef",
        dest="gdef",
        action="store_true",
        help="Also output GDEF table information",
    )

    parser.add_argument(
        "-i", "--index",
        dest="index", type=int, default=-1,
        help="Font index number (required for collections)",
    )

    parser.add_argument(
        "--no-lookups",
        dest="nolookups",
        action="store_true",
        help="Just list languages and features, don't unparse lookups",
    )
    parser.add_argument(
        "--config", default=None, help="config file to process", metavar="CONFIG"
    )
    parser.add_argument(
        "-O", "--optimize", dest="optimize", type=int, default=1, help="Run optimizer"
    )

    args = parser.parse_args(args)

    config = {}
    if args.config:
        import json

        with open(args.config) as f:
            config = json.load(f)

    if args.input.endswith("c") and args.index == -1:
        print("Must provide -i argument with TrueType/OpenType collection file", file=sys.stderr)
        sys.exit(1)

    font = TTFont(args.input, fontNumber=args.index)
    ff = unparse(font, do_gdef=args.gdef, doLookups=(not args.nolookups), config=config)
    # if args.optimize:
    #     Optimizer(ff).optimize(level=args.optimize)

    print(ff.asFea())

if __name__ == "__main__":
    main()


#!/usr/bin/env python
from fontFeatures.fontDameLib import unparse
from argparse import ArgumentParser

def main(args=None):
  parser = ArgumentParser()
  parser.add_argument("--config",default=None,
                      help="config file to process", metavar="CONFIG")
  parser.add_argument("input",
                      help="FontDame file to process", metavar="FILE")
  parser.add_argument("--font",
                      help="font file to process", metavar="FONT")
  parser.add_argument("--classfile",
                      help="class file to process", metavar="FEA")
  parser.add_argument("--optimize", dest="optimize", action='store_true',
                      help="Run optimizer")

  args = parser.parse_args(args)

  parsed = unparse(args.input, args.config, font= args.font)
  optlevel = 1
  if args.classfile:
    from fontTools.feaLib.parser import Parser
    parser = Parser(args.classfile)
    parser.parse()
    classes = parser.glyphclasses_.scopes_[0]
    parsed.scratch["glyphclasses"] = {}
    for k,v in classes.items():
      parsed.scratch["glyphclasses"][tuple(sorted(v.glyphs.glyphs))] = k
    parsed.scratch["index"] = 999

  # if args.optimize:
  # 	optlevel = 2
  # Optimizer(parsed).optimize(level=optlevel)

  print(parsed.asFea())

if __name__ == "__main__":
  main()


"""Routines for converting Chaining rules to and from XML."""



def _toXML(self, root):
    self._makeglyphslots(root, "input", self.input)
    return root


@classmethod
def fromXML(klass, el):
    """Creates a rule from a lxml Element object."""
    from fontFeatures import Routine, RoutineReference

    rule = klass(
        klass._slotArray(klass, el.find("input")),
        precontext=klass._slotArray(klass, el.find("precontext")),
        postcontext=klass._slotArray(klass, el.find("postcontext")),
        address=el.get("address"),
        languages=el.get("languages"),
        flags=int(el.get("flags") or 0),
    )
    lookupsxml = el.find("lookups")
    rule.lookups = []
    for slot in lookupsxml:
        routines = []
        for lu in slot:
            if lu.tag == "routinereference":
                routines.append(RoutineReference.fromXML(lu))
            elif lu.tag == "routine":
                routines.append(Routine.fromXML(lu))
        rule.lookups.append(routines)
    return rule



from collections import OrderedDict
from enum import IntEnum
import re
from youseedee import ucd_data


script_config = {
    "Invalid": dict(
        old_spec=False,
        virama=0,
        base_pos="last",
        reph_pos="before_post",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Devanagari": dict(
        old_spec=True,
        virama=0x094D,
        base_pos="last",
        reph_pos="before_post",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Bengali": dict(
        old_spec=True,
        virama=0x09CD,
        base_pos="last",
        reph_pos="after_sub",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Gurmukhi": dict(
        old_spec=True,
        virama=0x0A4D,
        base_pos="last",
        reph_pos="before_sub",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Gujarati": dict(
        old_spec=True,
        virama=0x0ACD,
        base_pos="last",
        reph_pos="before_post",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Oriya": dict(
        old_spec=True,
        virama=0x0B4D,
        base_pos="last",
        reph_pos="after_main",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Tamil": dict(
        old_spec=True,
        virama=0x0BCD,
        base_pos="last",
        reph_pos="after_post",
        reph_mode="implicit",
        blwf_mode="pre_and_post",
    ),
    "Telugu": dict(
        old_spec=True,
        virama=0x0C4D,
        base_pos="last",
        reph_pos="after_post",
        reph_mode="explicit",
        blwf_mode="post_only",
    ),
    "Kannada": dict(
        old_spec=True,
        virama=0x0CCD,
        base_pos="last",
        reph_pos="after_post",
        reph_mode="implicit",
        blwf_mode="post_only",
    ),
    "Malayalam": dict(
        old_spec=True,
        virama=0x0D4D,
        base_pos="last",
        reph_pos="after_main",
        reph_mode="log_repha",
        blwf_mode="pre_and_post",
    ),
    "Sinhala": dict(
        old_spec=False,
        virama=0x0DCA,
        base_pos="last_sinhala",
        reph_pos="after_post",
        reph_mode="explicit",
        blwf_mode="pre_and_post",
    ),
}

# Go from UCD categories to the ones we care about

syllabic_category_map = {
    "Other": "X",
    "Avagraha": "Symbol",
    "Bindu": "SM",
    "Brahmi_Joining_Number": "PLACEHOLDER",  # Don't care.
    "Cantillation_Mark": "A",
    "Consonant": "C",
    "Consonant_Dead": "C",
    "Consonant_Final": "CM",
    "Consonant_Head_Letter": "C",
    "Consonant_Killer": "M",  # U+17CD only.
    "Consonant_Medial": "CM",
    "Consonant_Placeholder": "PLACEHOLDER",
    "Consonant_Preceding_Repha": "Repha",
    "Consonant_Prefixed": "X",  # Don't care.
    "Consonant_Subjoined": "CM",
    "Consonant_Succeeding_Repha": "CM",
    "Consonant_With_Stacker": "CS",
    "Gemination_Mark": "SM",  # https://github.com/harfbuzz/harfbuzz/issues/552
    "Invisible_Stacker": "Coeng",
    "Joiner": "ZWJ",
    "Modifying_Letter": "X",
    "Non_Joiner": "ZWNJ",
    "Nukta": "N",
    "Number": "PLACEHOLDER",
    "Number_Joiner": "PLACEHOLDER",  # Don't care.
    "Pure_Killer": "M",  # Is like a vowel matra.
    "Register_Shifter": "RS",
    "Syllable_Modifier": "SM",
    "Tone_Letter": "X",
    "Tone_Mark": "N",
    "Virama": "H",
    "Visarga": "SM",
    "Vowel": "V",
    "Vowel_Dependent": "M",
    "Vowel_Independent": "V",
    # The following come from reassignment
    "Symbol": "Symbol",
    "Ra": "Ra",
    "Dotted_Circle": "DOTTEDCIRCLE",
}

## Set up the state machine for syllable matching and syllabic
## categories using composed regular expressions


states = OrderedDict(
    c="C|Ra",
    n="(ZWNJ?RS)?(N N?)?",
    z="ZWJ|ZWNJ",
    reph="Ra H|Repha",
    cn="c ZWJ?n?",
    forced_rakar="ZWJ H ZWJ Ra",
    symbol="Symbol N?",
    matra_group="z*M N?(H|forced_rakar)?",
    syllable_tail="(z?SM SM?ZWNJ?)?A*",
    halant_group="z?H(ZWJ N?)?",
    final_halant_group="halant_group|H ZWNJ",
    medial_group="CM?",
    halant_or_matra_group="final_halant_group|matra_group*",
    complex_syllable_tail="(halant_group cn)*medial_group halant_or_matra_group syllable_tail",
    consonant_syllable="((Repha|CS)?cn complex_syllable_tail)",
    vowel_syllable="(reph?V n?(complex_syllable_tail|ZWJ))",
    standalone_cluster="((Repha|CS)?PLACEHOLDER|reph?DOTTEDCIRCLE)n? complex_syllable_tail",
    symbol_cluster="symbol syllable_tail",
    broken_cluster="reph? n? complex_syllable_tail",
)


def make_syllable_machine(states, additional_categories=[]):
    categories = set(syllabic_category_map.values()) | set(additional_categories)
    syllable_machine = {}
    tail = r"\([^\)]+\)=\d+"
    for value in categories:
        syllable_machine[value] = f"<{value}>" + tail
    syllable_machine["other"] = r"<[^\>]+>" + tail

    for category, definition in states.items():
        # Recursively replace any known states with their definitions:
        states = (
            "\\b("
            + "|".join(sorted(syllable_machine.keys(), key=len, reverse=True))
            + ")\\b"
        )
        definition = re.sub(
            states, lambda match: "(" + syllable_machine[match[1]] + ")", definition
        )
        definition = definition.replace(" ", "")
        syllable_machine[category] = definition
    return syllable_machine


syllable_machine_indic = make_syllable_machine(states)


class IndicPosition(IntEnum):
    START = 0

    RA_TO_BECOME_REPH = 1
    PRE_M = 2
    PRE_C = 3

    BASE_C = 4
    AFTER_MAIN = 5

    ABOVE_C = 6

    BEFORE_SUB = 7
    BELOW_C = 8
    AFTER_SUB = 9

    BEFORE_POST = 10
    POST_C = 11
    AFTER_POST = 12

    FINAL_C = 13
    SMVD = 14

    END = 15


def is_ra(u):
    return u in [
        0x0930,
        0x09B0,
        0x09F0,
        0x0A30,
        0x0AB0,
        0x0B30,
        0x0BB0,
        0x0C30,
        0x0CB0,
        0x0D30,
        0x0DBB,
        0x179A,
    ]


def IndicPositionalCategory2IndicPosition(ipc):
    if ipc == "Left":
        return IndicPosition.PRE_C
    if ipc == "Top":
        return IndicPosition.ABOVE_C
    if ipc == "Bottom":
        return IndicPosition.BELOW_C

    if ipc == "Right":
        return IndicPosition.POST_C
    if ipc == "Bottom_And_Right":
        return IndicPosition.POST_C
    if ipc == "Left_And_Right":
        return IndicPosition.POST_C
    if ipc == "Top_And_Bottom_And_Left":
        return IndicPosition.POST_C

    if ipc == "Top_And_Bottom":
        return IndicPosition.BELOW_C
    if ipc == "Top_And_Bottom_And_Left":
        return IndicPosition.BELOW_C
    if ipc == "Top_And_Bottom_And_Right":
        return IndicPosition.POST_C

    if ipc == "Top_And_Left":
        return IndicPosition.ABOVE_C
    if ipc == "Top_And_Left_And_Right":
        return IndicPosition.POST_C
    if ipc == "Top_And_Right":
        return IndicPosition.POST_C

    if ipc == "Overstruck":
        return IndicPosition.AFTER_MAIN
    if ipc == "Visual_Order_Left":
        return IndicPosition.PRE_M

    return IndicPosition.END


matra_pos_left = {"Default": IndicPosition.PRE_M}
matra_pos_right = {
    # Telu and Knda need special cases
    "Devanagari": IndicPosition.AFTER_SUB,
    "Bengali": IndicPosition.AFTER_POST,
    "Gurmukhi": IndicPosition.AFTER_POST,
    "Gujarati": IndicPosition.AFTER_POST,
    "Oriya": IndicPosition.AFTER_POST,
    "Tamil": IndicPosition.AFTER_POST,
    "Malayalam": IndicPosition.AFTER_POST,
    "Default": IndicPosition.AFTER_SUB,
}
matra_pos_top = {
    "Devanagari": IndicPosition.AFTER_SUB,
    "Gurmukhi": IndicPosition.AFTER_POST,
    "Gujarati": IndicPosition.AFTER_SUB,
    "Oriya": IndicPosition.AFTER_MAIN,
    "Tamil": IndicPosition.AFTER_SUB,
    "Telugu": IndicPosition.BEFORE_SUB,
    "Kannada": IndicPosition.BEFORE_SUB,
    "Sinhala": IndicPosition.AFTER_SUB,
    "Default": IndicPosition.AFTER_SUB,
}
matra_pos_bottom = {
    "Devanagari": IndicPosition.AFTER_SUB,
    "Bengali": IndicPosition.AFTER_SUB,
    "Gurmukhi": IndicPosition.AFTER_POST,
    "Gujarati": IndicPosition.AFTER_POST,
    "Oriya": IndicPosition.AFTER_SUB,
    "Tamil": IndicPosition.AFTER_POST,
    "Telugu": IndicPosition.BEFORE_SUB,
    "Malayalam": IndicPosition.AFTER_POST,
    "Sinhala": IndicPosition.AFTER_SUB,
    "Default": IndicPosition.AFTER_SUB,
}


def set_matra_position(item):
    script = ucd_data(item.codepoint)["Script"]
    u = item.codepoint
    if item.syllabic_position == IndicPosition.PRE_C:
        selector = matra_pos_left
    elif item.syllabic_position == IndicPosition.POST_C:
        selector = matra_pos_right
        if script == "Telugu":
            if u <= 0x0C42:
                item.syllabic_position = IndicPosition.BEFORE_SUB
            else:
                item.syllabic_position = IndicPosition.AFTER_SUB
            return
        if script == "Kannada":
            if u < 0x0CC3 or u > 0xCD6:
                item.syllabic_position = IndicPosition.BEFORE_SUB
            else:
                item.syllabic_position = IndicPosition.AFTER_SUB
            return
    elif item.syllabic_position == IndicPosition.ABOVE_C:
        selector = matra_pos_top
    elif item.syllabic_position == IndicPosition.BELOW_C:
        selector = matra_pos_bottom
    else:
        return
    item.syllabic_position = selector.get(script, selector["Default"])


def reassign_category_and_position_indic(item):
    cp = item.codepoint
    if 0x0953 <= cp <= 0x0954:
        item.syllabic_category = "SM"
    elif 0x0A72 <= cp <= 0x0A73 or 0x1CF5 <= cp <= 0x1CF6:
        item.syllabic_category = "C"
    elif 0x1CE2 <= cp <= 0x1CE8 or cp == 0x1CED:
        item.syllabic_category = "A"
    elif 0xA8F2 <= cp <= 0xA8F7 or 0x1CE9 <= cp <= 0x1CEC or 0x1CEE <= cp <= 0x1CF1:
        item.syllabic_category = "Symbol"
    elif cp == 0x0A51:  # https://github.com/harfbuzz/harfbuzz/issues/524
        item.syllabic_category = "M"
        item.syllabic_position = IndicPosition.BELOW_C
    elif cp == 0x11301 or cp == 0x11303:
        item.syllabic_category = "SM"
    elif cp == 0x1133B or cp == 0x1133C:
        item.syllabic_category = "N"
    elif cp == 0x0AFB:
        item.syllabic_category = "N"
    elif cp == 0x0980 or cp == 0x09FC or cp == 0x0C80 or cp == 0x2010 or cp == 0x2011:
        item.syllabic_category = "PLACEHOLDER"
    elif cp == 0x25CC:
        item.syllabic_category = "DOTTEDCIRCLE"

    isc = item.syllabic_category

    is_medial = isc == "CM"
    is_consonant = (
        isc in ["C", "CS", "Ra", "V", "PLACEHOLDER", "DOTTEDCIRCLE"] or is_medial
    )

    if is_consonant:
        item.syllabic_position = IndicPosition.BASE_C
        if is_ra(cp):
            item.syllabic_category = "Ra"
    elif isc == "M":
        set_matra_position(item)
    elif isc in ["SM", "A", "Symbol"]:
        item.syllabic_position = IndicPosition.SMVD
    if cp == 0x0B01:
        item.syllabic_position = IndicPosition.BEFORE_SUB



def shaper_inputs(self):
    return self.input


def __find_masked_ix(buf, ix):
    # This probably could be done better...
    for i in range(len(buf.items)):
        if buf.mask[i] == ix:
            return i
    return -1


def _do_apply(self, buf, ix, namedclasses={}):
    from fontFeatures import RoutineReference

    # Save buffer mask
    flags = buf.flags
    markFilteringSet = buf.markFilteringSet
    markAttachmentSet = buf.markAttachmentSet

    if ix + len(self.lookups) - 1 > len(buf.mask):
        return

    old_unmasked_indexes = [buf.mask[ix + i] for i in range(len(self.lookups))]
    for i, lookups in enumerate(self.lookups):
        if not lookups:
            continue
        for routine in lookups:
            assert isinstance(routine, RoutineReference)
            routine = routine.routine
            # Adjust mask and recompute index?
            unmasked_ix = old_unmasked_indexes[i]
            for rule in routine.rules:
                buf.set_mask(
                    rule.flags, routine.markFilteringSet, routine.markAttachmentSet
                )
                newix = __find_masked_ix(buf, unmasked_ix)
                if rule.would_apply_at_position(
                    buf, newix, namedclasses
                ) and rule._do_apply(buf, newix):
                    break

    buf.set_mask(flags, markFilteringSet, markAttachmentSet)
    return len(self.input) - 1



#!/usr/bin/env python3
from fontFeatures.shaperLib.Buffer import Buffer
from fontFeatures.ttLib import unparse
from fontTools.ttLib import TTFont
from fontFeatures.shaperLib.Shaper import Shaper
import sys
from babelfont import load
from fontTools.ttLib import TTFont
import argparse
import logging
import re

def main(args=None):
    parser = argparse.ArgumentParser(description='Shape text')
    parser.add_argument('font', metavar='OTF',
                        help='Font file')
    parser.add_argument('--verbose', '-V', action="count", dest='logging',
                        help='Output interim shaping results')
    parser.add_argument("--ned", "--no-extra-data", dest="ned", action='store_true', help="No Extra Data; Do not output clusters or advances")
    parser.add_argument("--no-glyph-names", dest="ngn", action='store_true', help="Output glyph indices instead of names")
    parser.add_argument("--no-positions", dest="np", action='store_true', help="Do not output glyph positions")
    parser.add_argument("--additional", help='Additional information')
    parser.add_argument('--features', help='Feature string')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-u', help='Unicodes')
    group.add_argument('string', metavar='STRING',
                        help='Text to shape', nargs="?")

    args = parser.parse_args(args)

    logging.basicConfig(format='%(message)s')

    if args.logging:
        logging.getLogger("fontFeatures.shaperLib").setLevel(logging.INFO)
        if args.logging >= 2:
            logging.getLogger("fontFeatures.shaperLib").setLevel(logging.DEBUG)

    font = load(args.font)
    if args.font.endswith(".ttf") or args.font.endswith(".otf"):
        ff = unparse(TTFont(args.font))
    if args.u:
        args.u = re.sub("[uU]+|0x", "", args.u)
        splitup = re.split(r"[\s,]", args.u)
        args.string = "".join([chr(int(x,16)) for x in splitup])

    buf = Buffer(font, unicodes=args.string)
    shaper = Shaper(ff, font)
    if args.features:
        shaper.execute(buf, features=args.features)
    else:
        shaper.execute(buf)
    serialize_options = {}
    if args.ngn:
        serialize_options["names"] = False
    if args.ned:
        serialize_options["ned"] = True
    if args.np:
        serialize_options["position"] = False
    if args.additional:
        serialize_options["additional"] = args.additional
    if buf.direction == "RTL":
        buf.items = list(reversed(buf.items))
    print(buf.serialize(**serialize_options))

if __name__ == "__main__":
    main()


"""Routines for converting Substitution rules to and from XML."""
from fontFeatures import Routine


def _toXML(self, root):
    self._makeglyphslots(root, "from", self.input)
    self._makeglyphslots(root, "to", self.replacement)
    return root


@classmethod
def fromXML(klass, el):
    """Creates a rule from a lxml Element object."""
    rule = klass(
        klass._slotArray(klass, el.find("from")),
        klass._slotArray(klass, el.find("to")),
        precontext=klass._slotArray(klass, el.find("precontext")),
        postcontext=klass._slotArray(klass, el.find("postcontext")),
        address=el.get("address"),
        languages=el.get("languages"),
        reverse=el.get("reverse"),
        flags=el.get("flags"),
    )
    if el.find("lookups"):
        rule.lookups = []
        for slot in list(el):
            routines = [Routine.fromXML(x) for x in slot.findall("routine")]
            rule.lookups.append(routines)
    return rule



# Code for converting a Attachment object into feaLib statements
import fontTools.feaLib.ast as feaast
from fontFeatures.utils import categorize_glyph


def fix_scalar(scalar):
    if isinstance(scalar, int):
        return scalar
    if isinstance(scalar, float):
        return int(scalar)
    if hasattr(scalar, "axes"):
        # assume it's a variable scalar of some kind
        return scalar
    raise ValueError("Illegal anchor position %s" % scalar)


def _glyphref(g):
    if len(g) == 1:
        return feaast.GlyphName(g[0])
    return feaast.GlyphClass([feaast.GlyphName(x) for x in g])


def sortByAnchor(self):
    anchors = {}
    for name, pos in self.marks.items():
        if pos not in anchors:
            anchors[pos] = []
        anchors[pos].append(name)
    self.markslist = [(v, k) for k, v in anchors.items()]

    anchors = {}
    for name, pos in self.bases.items():
        if pos not in anchors:
            anchors[pos] = []
        anchors[pos].append(name)
    self.baseslist = [(v, k) for k, v in anchors.items()]


def feaPreamble(self, ff):
    if self.is_cursive:
        return []
    sortByAnchor(self)
    if "mark_classes_done" not in ff.scratch:
        ff.scratch["mark_classes_done"] = {}
    b = feaast.Block()
    for mark in self.markslist:
        if (self.fullname, tuple(mark[0])) not in ff.scratch["mark_classes_done"]:
            b.statements.append(
                feaast.MarkClassDefinition(
                    feaast.MarkClass(self.fullname),
                    feaast.Anchor(fix_scalar(mark[1][0]), fix_scalar(mark[1][1])),
                    _glyphref(mark[0]),
                )
            )
            ff.scratch["mark_classes_done"][(self.fullname, tuple(mark[0]))] = True
    return [b]


def asFeaAST(self):
    b = feaast.Block()
    if self.is_cursive:
        allglyphs = set(self.bases.keys()) | set(self.marks.keys())
        for g in allglyphs:
            b.statements.append(
                feaast.CursivePosStatement(
                    _glyphref([g]),
                    g in self.bases
                    and feaast.Anchor(
                        fix_scalar(self.bases[g][0]), fix_scalar(self.bases[g][1])
                    ),
                    g in self.marks
                    and feaast.Anchor(
                        fix_scalar(self.marks[g][0]), fix_scalar(self.marks[g][1])
                    ),
                )
            )
    else:
        if not hasattr(self, "baseslist"):
            sortByAnchor(self)  # e.g. when testing
        for base in self.baseslist:
            statementtype = feaast.MarkBasePosStatement
            if self.font:
                if categorize_glyph(self.font, base[0][0])[0] == "mark":
                    statementtype = feaast.MarkMarkPosStatement
            if self.force_markmark:
                statementtype = feaast.MarkMarkPosStatement
            b.statements.append(
                statementtype(
                    _glyphref(base[0]),
                    [
                        [
                            feaast.Anchor(
                                fix_scalar(base[1][0]), fix_scalar(base[1][1])
                            ),
                            feaast.MarkClass(self.fullname),
                        ]
                    ],
                )
            )

    return b



import logging

__all__ = [
    "would_apply_at_position",
    "pre_post_context_matches",
    "_expand_slot",
]


def i2s(buffer_items):
    return "|".join([x.glyph for x in buffer_items])


def _expand_slot(slot, namedclasses={}):
    expanded = []
    for g in slot:
        if g.startswith("@"):
            expanded.extend(namedclasses.get(g[1:], ""))
        else:
            expanded.append(g)
    return expanded


def glyphs_match(buffer_glyphs, routine_glyphs, namedclasses={}):
    if len(buffer_glyphs) != len(routine_glyphs):
        return False
    for a, b in zip(buffer_glyphs, routine_glyphs):
        if a.glyph not in _expand_slot(b, namedclasses):
            return False
    return True


def pre_post_context_matches(self, buf, ix, namedclasses={}):
    if hasattr(self, "precontext") and self.precontext:
        if ix < len(self.precontext):
            logging.getLogger("fontFeatures.shaperLib").debug(
                " - No, not enough precontext"
            )
            return False
        precontext = buf[ix - len(self.precontext) : ix]
        if not glyphs_match(precontext, self.precontext, namedclasses):
            logging.getLogger("fontFeatures.shaperLib").debug(
                " - No, precontext doesn't match %s != %s"
                % (i2s(precontext), self.precontext)
            )
            return False
    if hasattr(self, "postcontext") and self.postcontext:
        coverage = self.shaper_inputs()
        coverage_l = len(coverage)
        end_of_coverage = ix + coverage_l
        if end_of_coverage + len(self.postcontext) > len(buf):
            logging.getLogger("fontFeatures.shaperLib").debug(
                " - No, not enough postcontext"
            )
            return False
        postcontext = buf[end_of_coverage : end_of_coverage + len(self.postcontext)]
        if not glyphs_match(postcontext, self.postcontext, namedclasses):
            logging.getLogger("fontFeatures.shaperLib").debug(
                " - No, postcontext doesn't match %s != %s"
                % (i2s(postcontext), self.postcontext)
            )
            return False
    return True


def would_apply_at_position(self, buf, ix, namedclasses={}):
    logging.getLogger("fontFeatures.shaperLib").debug(
        "Testing if rule would apply at position %i" % (ix)
    )
    coverage = self.shaper_inputs()
    coverage_l = len(coverage)
    if coverage_l < 1:
        return False
    buffer_glyphs = buf[ix : ix + coverage_l]

    if not glyphs_match(buffer_glyphs, coverage, namedclasses):
        logging.getLogger("fontFeatures.shaperLib").debug(
            " - No! %s != %s" % (i2s(buffer_glyphs), coverage)
        )
        return False

    if not pre_post_context_matches(self, buf, ix, namedclasses):
        return False

    logging.getLogger("fontFeatures.shaperLib").debug(
        " - Yes! %s == %s" % (i2s(buffer_glyphs), coverage)
    )
    return True



# Port of glyphtools.categorize_glyph, isolated here to reduce
# dependencies.


def categorize_glyph(font, glyphname):
    """Return the category of the given glyph.

    Args:
        font: a ``fontTools`` TTFont object OR a ``glyphsLib`` GSFontMaster
            object OR a ``babelfont`` Font object.
        glyphname: name of the glyph.

    Returns:
        A two-element tuple. The first element is one of the following
        strings: ``unknown``, ``base``, ``mark``, ``ligature``, ``component``.
        If the glyph is a mark, the second element is the mark attachment
        class number.
    """
    if hasattr(font, "userData"):  # Glyphs
        c = font.font.glyphs[glyphname].category
        sc = font.font.glyphs[glyphname].subCategory
        if sc == "Ligature":
            return ("ligature", None)
        if c == "Mark":
            return ("mark", None)
        return ("base", None)
    if hasattr(font, "_set_kerning"):  # Babelfont 1/2/3
        if hasattr(font, "_master_map") or hasattr(font, "get_glyph_layer"):
            return (font.glyphs[glyphname].category, None)  # Babelfont 3
        else:
            return (font[glyphname].category, None)  # Babelfont 1/2

    # TT font
    if "GDEF" not in font:
        return ("unknown", None)
    gdef = font["GDEF"].table
    classdefs = gdef.GlyphClassDef.classDefs
    if glyphname not in classdefs:
        return ("unknown", None)
    if classdefs[glyphname] == 1:
        return ("base", None)
    if classdefs[glyphname] == 2:
        return ("ligature", None)
    if classdefs[glyphname] == 3:
        # Now find attachment class
        mclass = None
        if gdef.MarkAttachClassDef:
            classdef = gdef.MarkAttachClassDef.classDefs
            if glyphname in classdef:
                mclass = classdef[glyphname]
        return ("mark", mclass)
    if classdefs[glyphname] == 4:
        return ("component", None)
    return ("unknown", None)
